;;; bbdb-merge.el --- merge and sync framework for BBDB

;; Copyright (C) 2000 Waider
;; Copyright (C) 2010 Roland Winkler <roland.winkler@physik.uni-erlangen.de>

;; This file is part of the Insidious Big Brother Database (aka BBDB),

;; BBDB is free software: you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 2 of the License, or
;; (at your option) any later version.

;; BBDB is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs.  If not, see <http://www.gnu.org/licenses/>.

;;; Commentary:
;;; This file provides a merge and sync framework for BBDB.
;;; See bbdb.texinfo for documentation.

(require 'bbdb)
(require 'bbdb-com)

;;; to do:
;;; smarter phone, notes and address merging.

;;;###autoload
(defun bbdb-merge-record (new-record &optional merge-record override)
  "Generic merge function.

Merges NEW-RECORD into bbdb, using DATE to check who's more
up-to-date and OVERRIDE to decide who gets precedence if two dates
match. DATE can be extracted from a notes if it's an alist with an
element marked timestamp. Set OVERRIDE to 'new to allow the new record
to stomp on existing data, 'old to preserve existing data or nil to
merge both together. If it cannot find a record to merge with, it will
create a new record. If MERGE-RECORD is set, it's a record discovered
by other means that should be merged with.

Returns the Grand Unified Record."

  (let* ((firstname (bbdb-record-firstname new-record))
         (lastname (bbdb-record-lastname new-record))
         (aka (bbdb-record-aka new-record))
         (mails (bbdb-record-mail new-record))
         (addresses (bbdb-record-address new-record))
         (phones (bbdb-record-phone new-record))
         (organization (bbdb-record-organization new-record))
         (notes (bbdb-record-notes new-record))
         (name (bbdb-record-name new-record))
         (date (cdr (assq 'timestamp notes)))
         olddate)

    ;; See if we have a record that looks right, using an intertwingle
    ;; search. Could probably parameterize that.
    ;; bbdb-merge-search-function or some such.
    ;; FIXME: if we have two records to merge we duplicate the code
    ;; in `bbdb-merge-records-internal'
    (unless merge-record
      (setq merge-record (car (bbdb-message-search name mails))))

    (if merge-record
        (progn
          ;; if date is unset, set it to the existing record's date.
          (setq olddate (bbdb-record-note merge-record 'timestamp)
                date (or date olddate))
          ;; FIXME if date & olddate are STILL unset, set to today's date.

          ;; if the old record is actually newer, invert the sense of override
          (if (string-lessp olddate date)
              (setq override (cond ((eq 'old override) 'new)
                                   ((eq 'new override) 'old))))

          (bbdb-record-set-firstname merge-record
           (if (null override)
               (bbdb-merge-strings (bbdb-record-firstname merge-record)
                                   firstname " ")
             (if (eq 'new override) firstname
               (bbdb-record-firstname merge-record))))

          (bbdb-record-set-lastname merge-record
           (if (null override)
               (bbdb-merge-strings (bbdb-record-lastname merge-record)
                                   lastname " ")
             (if (eq 'new override) lastname
               (bbdb-record-lastname merge-record))))

          (bbdb-record-set-organization merge-record
           (if (null override)
               (bbdb-merge-strings (bbdb-record-organization merge-record)
                                   organization " ")
             (if (eq 'new override) organization
               (bbdb-record-organization merge-record))))

          (bbdb-record-set-aka
           merge-record
           (if (null override)
               (bbdb-merge-lists
                (bbdb-record-aka merge-record)
                (if (listp aka) aka (list aka)) 'string= 'downcase)
             (if (eq 'new override) aka
               (bbdb-record-aka merge-record))))

          (bbdb-record-set-mail
           merge-record
           (if (null override)
               (bbdb-merge-lists
                (bbdb-record-mail merge-record) mails 'string= 'downcase)
             (if (eq 'new override) mails
               (bbdb-record-mail merge-record))))

          (bbdb-record-set-phone
           merge-record
           (if (null override)
               (bbdb-merge-lists
                (bbdb-record-phone merge-record) phones 'equal)
             (if (eq 'new override) phones
               (bbdb-record-phone merge-record))))

          (bbdb-record-set-address
           merge-record
           (if (null override)
               (bbdb-merge-lists
                (bbdb-record-address merge-record) addresses 'equal)
             (if (eq 'new override) addresses
               (bbdb-record-address merge-record))))

          ;; lifted from bbdb-com.el
          (let ((n1 (bbdb-record-notes merge-record))
                (n2 notes)
                tmp
                (bbdb-merge-notes-function ;; XXX
                 'bbdb-merge-strings))
            (or (equal n1 n2)
                (progn
                  (or (listp n1) (setq n1 (list (cons 'notes n1))))
                  (or (listp n2) (setq n2 (list (cons 'notes n2))))
                  (while n2
                    (if (setq tmp (assq (car (car n2)) n1))
                        (setcdr tmp
                                (funcall (or (cdr (assq (car (car n2))
                                                        bbdb-merge-notes-function-alist))
                                             bbdb-merge-notes-function)
                                         (cdr tmp) (cdr (car n2))))
                      (setq n1 (nconc n1 (list (car n2)))))
                    (setq n2 (cdr n2)))
                  (bbdb-record-set-notes merge-record n1)))))

      ;; we could not find a record, so create one
      (setq merge-record
            (bbdb-create-internal name nil nil organization mails phones addresses notes))
      ;; bite me, bbdb-create-internal
      (bbdb-record-set-firstname merge-record firstname)
      (bbdb-record-set-lastname merge-record lastname))

    ;; more general bitingness
    (if (equal (bbdb-record-firstname merge-record) "")
        (bbdb-record-set-firstname merge-record nil))
    (if (equal (bbdb-record-lastname merge-record) "")
        (bbdb-record-set-lastname merge-record nil))

    ;; fix up the in-memory copy.
    (bbdb-change-record merge-record t t)
    (let ((name (bbdb-record-name merge-record)))
      (if (> (length name) 0)
          (bbdb-remhash name merge-record)))
    (dolist (organization (bbdb-record-organization merge-record))
      (bbdb-remhash organization merge-record))
    (bbdb-record-unset-name merge-record)
    (if (or (bbdb-record-lastname merge-record)
            (bbdb-record-firstname merge-record))
        (bbdb-puthash (bbdb-record-name merge-record) merge-record))
    (if (bbdb-record-organization merge-record)
        (bbdb-puthash (bbdb-record-organization merge-record) merge-record))
    (unless (memq merge-record bbdb-changed-records)
      (push merge-record bbdb-changed-records))
    merge-record))

(defun bbdb-instring( s1 s2 )
  "Return t if string S1 contains string S2."
;;  (and case-fold-search
;;       (setq s1 (downcase s1)
;;             s2 (downcase s2)))
  (catch 'done
    (while (>= (length s1) (length s2))
      (if (string= s2 (substring s1 0 (length s2)))
          (throw 'done t)
        (setq s1 (substring s1 1))))
    (throw 'done nil)))

(defun bbdb-merge-strings (s1 s2 &optional sep)
  "Merge two strings S1 and S2 uniquely.
If S1 does not contain S2, return s1+sep+s2."
  ;; This treats S1 and S2 aymmetrically.
  (cond ((or (null s1) (string-equal s1 "")) s2)
        ((or (null s2) (string-equal s2 "")) s1)
        (t (if (bbdb-instring s2 s1) s1
             (concat s1 (or sep "") s2)))))

;;;###autoload
(defun bbdb-merge-file (&optional bbdb-new override match-fun)
  "Merge a bbdb file into the in-core bbdb."
  (interactive "fMerge bbdb file: ")
  (or bbdb-silent-internal
      bbdb-silent
      (message "Merging %s" bbdb-new))
  ;; argh urgle private environment
  (let* ((bbdb-live-file bbdb-file)
         (bbdb-file bbdb-new)
         (bbdb-live-buffer-name bbdb-buffer-name)
         (bbdb-buffer-name "*BBDB-merge*")
         bbdb-buffer ;; hack hack
         (new-records (bbdb-records))
         bbdb-buffer ;; hack hack
         (bbdb-file bbdb-live-file)
         (bbdb-buffer-name bbdb-live-buffer-name)
         (bbdb-merge-notes-function 'bbdb-merge-strings))

    ;; merge everything
    (mapc (lambda (record)
            (bbdb-merge-record record
                               (and match-fun
                                    (funcall match-fun record))
                               override))
            new-records))
  ;; hack
  (setq bbdb-buffer (or (get-file-buffer bbdb-file) nil)))

(defun bbdb-add-or-update-phone (record label string)
  "Add or update a phone number in the current record.

Insert into RECORD phone number STRING for LABEL.
This will automatically overwrite an existing phone field for the same label."
  (let* ((phone (make-vector (if (eq bbdb-phone-style 'nanp)
                                 bbdb-phone-length
                               2)
                             nil)))
    (if (= 2 (length phone))
        (aset phone 1 string)
      (let ((newp (bbdb-parse-phone string)))
        (bbdb-phone-set-area phone (nth 0 newp))
        (bbdb-phone-set-exchange phone (nth 1 newp))
        (bbdb-phone-set-suffix phone (nth 2 newp))
        (bbdb-phone-set-extension phone (or (nth 3 newp) 0))))
    (bbdb-phone-set-label phone label)

    ;; "phone" now contains a suitable record
    ;; we need to check if this is already in the phones list
    (let ((phones (bbdb-record-phone record))
          phones-list)
      (setq phones-list phones)
      (while (car phones-list)
        (if (string= (bbdb-phone-label (car phones-list))
                     label)
            (setq phones (delete (car phones-list) phones)))
        (setq phones-list (cdr phones-list)))


      (bbdb-record-set-phone record
                              (nconc phones (list phone))))
    (bbdb-change-record record)

    ;; update display if record is visible
    (and (get-buffer-window bbdb-buffer-name)
         (bbdb-display-records (list record)))
    nil))

(provide 'bbdb-merge)
